#cython: language_level=3
#distutils: define_macros=CYTHON_TRACE_NOGIL=1

from .parser cimport (
    ParseInfo,
    line_number_strings as _line_number_strings,
    advance_to_non_space as _advance_to_non_space,
    get_slashed_char as _get_slashed_char,
    parse_unquoted_plist_string as _parse_unquoted_plist_string,
    parse_plist_string as _parse_plist_string,
)
from .util cimport (
    tounicode,
    is_valid_unquoted_string_char as _is_valid_unquoted_string_char,
)
from .writer cimport string_needs_quotes
from cpython.mem cimport PyMem_Free
from cpython.unicode cimport (
    PyUnicode_AsUCS4Copy, PyUnicode_GET_LENGTH,
)


cdef class ParseContext:

    cdef unicode s
    cdef ParseInfo pi
    cdef Py_UCS4 *buf
    cdef object dict_type

    def __cinit__(
        self,
        string,
        Py_ssize_t offset=0,
        dict_type=dict,
        bint use_numbers=False
    ):
        self.s = tounicode(string)
        cdef Py_ssize_t length = PyUnicode_GET_LENGTH(self.s)
        self.buf = PyUnicode_AsUCS4Copy(self.s)
        if not self.buf:
            raise MemoryError()
        self.dict_type = dict_type
        self.pi = ParseInfo(
            begin=self.buf,
            curr=self.buf + offset,
            end=self.buf + length,
            dict_type=<void*>dict_type,
            use_numbers=use_numbers,
        )

    def __dealloc__(self):
        PyMem_Free(self.buf)


def is_valid_unquoted_string_char(Py_UCS4 c):
    return _is_valid_unquoted_string_char(c)


def line_number_strings(s, offset=0):
    cdef ParseContext ctx = ParseContext(s, offset)
    return _line_number_strings(&ctx.pi)


def advance_to_non_space(s, offset=0):
    cdef ParseContext ctx = ParseContext(s, offset)
    eof = not _advance_to_non_space(&ctx.pi)
    return None if eof else s[ctx.pi.curr - ctx.pi.begin]


def get_slashed_char(s, offset=0):
    cdef ParseContext ctx = ParseContext(s, offset)
    return _get_slashed_char(&ctx.pi)


def parse_unquoted_plist_string(s):
    cdef ParseContext ctx = ParseContext(s)
    return _parse_unquoted_plist_string(&ctx.pi)


def parse_plist_string(s, required=True):
    cdef ParseContext ctx = ParseContext(s)
    return _parse_plist_string(&ctx.pi, required=required)



#!/usr/bin/env python
from __future__ import absolute_import, unicode_literals
import argparse
import openstep_plist
import json
import binascii
import pydoc
from functools import partial
from io import StringIO, open


class BytesEncoder(json.JSONEncoder):
    def default(self, obj):
        from glyphsLib.types import BinaryData

        if isinstance(obj, (bytes, BinaryData)):
            return "<%s>" % binascii.hexlify(obj).decode()
        return json.JSONEncoder.default(self, obj)


def main(args=None):
    if args is None:
        import sys

        args = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument("infile", help="input file")
    parser.add_argument("outfile", help="output file", default="-", nargs="?")
    parser.add_argument(
        "-g", "--glyphs", help="use glyphsLib parser/writer", action="store_true"
    )
    parser.add_argument(
        "--no-pager", dest="pager", help="do not use pager", action="store_false"
    )
    parser.add_argument(
        "-j", "--json", help="use json to serialize", action="store_true", default=False
    )
    parser.add_argument("-i", "--indent", help="indentation level", type=int, default=2)
    args = parser.parse_args(args)

    if not args.glyphs:
        parse = partial(openstep_plist.load, use_numbers=True)
    else:

        def parse(fp, dict_type=dict):
            from glyphsLib.parser import Parser

            s = fp.read()
            p = Parser(current_type=dict_type)
            return p.parse(s)

    if args.json:
        dump = partial(
            json.dump, cls=BytesEncoder, sort_keys=True, indent=" " * args.indent
        )
    else:
        if args.glyphs:
            from glyphsLib.writer import dump
        else:
            dump = partial(openstep_plist.dump, indent=args.indent)

    with open(args.infile, "r", encoding="utf-8") as fp:
        data = parse(fp)

    if args.outfile == "-":
        if args.pager:
            buf = StringIO()
            dump(data, buf)
            pydoc.pager(buf.getvalue())
        else:
            dump(data, sys.stdout)
    else:
        with open(args.outfile, "w", encoding="utf-8") as fp:
            dump(data, fp)


if __name__ == "__main__":
    main()



from .parser import load, loads, ParseError
from .writer import dump, dumps

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "0.0.0+unknown"


__all__ = ["load", "loads", "dump", "dumps", "ParseError"]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import sys
from io import StringIO, BytesIO
from collections import OrderedDict
from openstep_plist._test import (
    line_number_strings,
    is_valid_unquoted_string_char,
    advance_to_non_space,
    get_slashed_char,
    parse_unquoted_plist_string,
    parse_plist_string,
)
import openstep_plist
from openstep_plist.parser import string_to_number

import pytest


def test_line_number_strings():
    assert line_number_strings("", 0) == 1
    assert line_number_strings("a\na", 1) == 1
    assert line_number_strings("a\na", 2) == 2
    assert line_number_strings("a\naa\n", 4) == 2
    assert line_number_strings("a\naa\na", 5) == 3
    assert line_number_strings("a\raa\ra", 5) == 3
    assert line_number_strings("a\r\naa\ra", 6) == 3
    assert line_number_strings("a\n\naa\n\na", 7) == 5


def test_is_valid_unquoted_string_char():
    assert is_valid_unquoted_string_char(ord("a"))
    assert is_valid_unquoted_string_char(ord("b"))
    assert is_valid_unquoted_string_char(ord("z"))
    assert is_valid_unquoted_string_char(ord("A"))
    assert is_valid_unquoted_string_char(ord("B"))
    assert is_valid_unquoted_string_char(ord("Z"))
    assert is_valid_unquoted_string_char(ord("_"))
    assert is_valid_unquoted_string_char(ord("$"))
    assert is_valid_unquoted_string_char(ord("/"))
    assert is_valid_unquoted_string_char(ord(":"))
    assert is_valid_unquoted_string_char(ord("."))
    assert is_valid_unquoted_string_char(ord("-"))
    assert not is_valid_unquoted_string_char(ord('"'))
    assert not is_valid_unquoted_string_char(ord(","))
    assert not is_valid_unquoted_string_char(ord("{"))
    assert not is_valid_unquoted_string_char(ord(")"))
    assert not is_valid_unquoted_string_char(ord(";"))
    assert not is_valid_unquoted_string_char(0x00)  # NULL
    assert not is_valid_unquoted_string_char(0x0A)  # \n
    assert not is_valid_unquoted_string_char(0x0D)  # \r


@pytest.mark.parametrize(
    "string, offset, expected",
    [
        ("", 0, None),
        (" a", 0, "a"),
        (" a", 1, "a"),
        (" a", 2, None),
        ("\t\ta", 1, "a"),
        ("\t\ta", 2, "a"),
        ("\t\ta", 3, None),
        ("abc//this is an inline comment", 3, None),
        ("abc //also this\n", 3, None),
        ("abc //this as well\n\nz", 3, "z"),
        ("abc/this is not a comment", 3, "/"),
        ("abc/", 3, "/"),  # not a comment either
        ("abcd /* C-style comments! */z", 4, "z"),
    ],
)
def test_advance_to_non_space(string, offset, expected):
    assert advance_to_non_space(string, offset) == expected


@pytest.mark.parametrize(
    "string, expected",
    [
        ("000", "\x00"),
        ("001", "\x01"),
        ("002", "\x02"),
        ("003", "\x03"),
        ("004", "\x04"),
        ("005", "\x05"),
        ("006", "\x06"),
        ("007", "\x07"),
        ("012", "\n"),
        ("111", "I"),
        ("111", "I"),
        ("200", "\xa0"),
        ("201", "\xc0"),
        ("375", "\xff"),
        ("376", "\ufffd"),
        ("376", "\ufffd"),
        ("U0000", "\u0000"),
        ("U0001", "\u0001"),
        ("U0411", "\u0411"),
        ("U00FA", "\u00fa"),
        ("a", "\a"),
        ("b", "\b"),
        ("f", "\f"),
        ("n", "\n"),
        ("r", "\r"),
        ("t", "\t"),
        ("v", "\v"),
        ('"', '"'),
        ("\n", "\n"),
        ("\\", "\\"),
        ("z", "z"),
    ],
)
def test_get_slashed_char(string, expected):
    assert get_slashed_char(string) == expected


@pytest.mark.parametrize(
    "string, expected",
    [
        ("a", "a"),
        ("abc;", "abc"),  # trailing chars left in buffer
        ("1", "1"),
        ("123456789", "123456789"),
        ("1.23456789", "1.23456789"),
    ],
)
def test_parse_unquoted_plist_string(string, expected):
    assert parse_unquoted_plist_string(string) == expected


def test_parse_unquoted_plist_string_EOF():
    with pytest.raises(openstep_plist.ParseError, match="Unexpected EOF"):
        parse_unquoted_plist_string("") == expected


@pytest.mark.parametrize(
    "string, expected",
    [
        ("a", "a"),
        ('"a"', "a"),
        ("'a'", "a"),
        ('"a\\012b"', ("a\nb")),
        # surrogate pair gets decoded as a single scalar value
        ('"\\UD83D\\UDCA9"', "\U0001F4A9"),  # '💩'
        # surrogate that don't go in pairs are simply passed through
        ('"\\UD83D"', "\ud83d"),
        ('"\\UD83D\\012"', "\ud83d\n"),
        ('"\\UDCA9"', "\udca9"),
        ('"\\UDCA9\\012"', "\udca9\n"),
    ],
)
def test_parse_plist_string(string, expected):
    assert parse_plist_string(string) == expected


def test_parse_plist_string_EOF():
    with pytest.raises(openstep_plist.ParseError, match="Unexpected EOF"):
        parse_plist_string("")
    with pytest.raises(openstep_plist.ParseError, match="Unterminated quoted string"):
        parse_plist_string("'a")


def test_parse_plist_string_invalid_char():
    with pytest.raises(openstep_plist.ParseError, match="Invalid string character"):
        parse_plist_string("\\")
    assert parse_plist_string("\\", required=False) is None


def test_parse_plist_array():
    assert openstep_plist.loads("(1)") == ["1"]
    assert openstep_plist.loads("(1,)") == ["1"]
    assert openstep_plist.loads("(\t1  \r\n, 2.2, c,\n)") == ["1", "2.2", "c"]
    assert openstep_plist.loads("('1', '2')") == ["1", "2"]
    assert openstep_plist.loads("(\n1,\n\"'2'\"\n)") == ["1", "'2'"]


@pytest.mark.parametrize("string, lineno", [("(a ", 1), ("(a,\nb,\r\nc", 3)])
def test_parse_plist_array_missing_comma(string, lineno):
    msg = "Missing ',' for array at line %d" % lineno
    with pytest.raises(openstep_plist.ParseError, match=msg):
        openstep_plist.loads(string)


@pytest.mark.parametrize("string, lineno", [("(a,", 1), ("(a,\nb, }", 2)])
def test_parse_plist_array_missing_paren(string, lineno):
    msg = r"Expected terminating '\)' for array at line %d" % lineno
    with pytest.raises(openstep_plist.ParseError, match=msg):
        openstep_plist.loads(string)


def test_parse_plist_array_empty():
    assert openstep_plist.loads("()") == []


def test_parse_plist_dict_empty():
    assert openstep_plist.loads("") == {}
    assert openstep_plist.loads("{}") == {}


@pytest.mark.parametrize(
    "string, expected",
    [
        ("{a=1;}", {"a": "1"}),
        ('{"a"="1";}', {"a": "1"}),
        ("{'a'='1';}", {"a": "1"}),
        ("{\na = 1;\n}", {"a": "1"}),
        ("{\na\n=\n1;\n}", {"a": "1"}),
        ("{a=1;b;}", {"a": "1", "b": "b"}),
    ],
)
def test_parse_plist_dict(string, expected):
    assert openstep_plist.loads(string) == expected


def test_parse_plist_dict_invalid():
    msg = "Unexpected character after key at line 1: u?','"
    with pytest.raises(openstep_plist.ParseError, match=msg):
        openstep_plist.loads("{a,}")

    msg = "Missing ';' on line 1"
    with pytest.raises(openstep_plist.ParseError, match=msg):
        openstep_plist.loads("{b ")

    msg = "Missing ';' on line 2"
    with pytest.raises(openstep_plist.ParseError, match=msg):
        openstep_plist.loads("{b = zzz;\nc = xxx}")

    msg = "Expected terminating '}' for dictionary at line 3"
    with pytest.raises(openstep_plist.ParseError, match=msg):
        openstep_plist.loads("{b = zzz;\nc = xxx;\nd = jjj;")


@pytest.mark.parametrize(
    "string, expected",
    [
        ("<AA>", b"\xaa"),
        ("<B1B0AFBA>", b"\xb1\xb0\xaf\xba"),
        ("<AA BB>", b"\xaa\xbb"),
        ("<cdef>", b"\xcd\xef"),
        ("<4142\n4344>", b"ABCD"),
    ],
)
def test_parse_plist_data(string, expected):
    assert openstep_plist.loads(string) == expected


def test_parse_plist_data_invalid():
    with pytest.raises(openstep_plist.ParseError, match="Expected terminating '>'"):
        openstep_plist.loads("<FF")

    msg = "Malformed data byte group at line 1: invalid hex digit: u?'Z'"
    with pytest.raises(openstep_plist.ParseError, match=msg):
        openstep_plist.loads("<Z")
    with pytest.raises(openstep_plist.ParseError, match=msg):
        openstep_plist.loads("<AZ")

    msg = "Malformed data byte group at line 1: uneven length"
    with pytest.raises(openstep_plist.ParseError, match=msg):
        openstep_plist.loads("<AAA")
    with pytest.raises(openstep_plist.ParseError, match=msg):
        openstep_plist.loads("<AAA>")


def test_parse_plist_object_invalid():
    with pytest.raises(openstep_plist.ParseError, match="Unexpected character"):
        openstep_plist.loads(";")
    with pytest.raises(
        openstep_plist.ParseError, match="Unexpected EOF while parsing plist"
    ):
        openstep_plist.loads("{a=")
    with pytest.raises(openstep_plist.ParseError, match="Junk after plist at line 3"):
        openstep_plist.loads("{a=1;\nb=2;\n}...")


def test_parse_string_resources():
    assert openstep_plist.loads("a=1;\n'b' = 2.4;\n'c' = \"hello world\";") == {
        "a": "1",
        "b": "2.4",
        "c": "hello world",
    }


def test_load():
    fp = StringIO("{a=1;}")
    assert openstep_plist.load(fp) == {"a": "1"}


def test_load_from_bytes():
    if sys.version_info.major < 3:
        assert openstep_plist.loads(b"{a=1;}") == {"a": "1"}
    else:
        with pytest.raises(TypeError, match="Could not convert to unicode"):
            openstep_plist.loads(b"{a=1;}")


@pytest.mark.parametrize(
    "string, expected",
    [
        ("{a = 2;}", {"a": 2}),
        ("{a = {b = -2;};}", {"a": {"b": -2}}),
        ("{a = (1.5, -23.9999);}", {"a": [1.5, -23.9999]}),
        ("{a = x123; b = -c; minus = -;}", {"a": "x123", "b": "-c", "minus": "-"}),
    ],
)
def test_loads_use_numbers(string, expected):
    assert openstep_plist.loads(string, use_numbers=True) == expected


def test_loads_dict_type():
    assert openstep_plist.loads(
        "{z = (a, b); y = (c, d); a = 'hello world';}", dict_type=OrderedDict
    ) == (OrderedDict([("z", ["a", "b"]), ("y", ["c", "d"]), ("a", "hello world")]))


@pytest.mark.parametrize(
    "string, expected",
    [("2", 2), ("-2", -2), ("1.5", 1.5), ("-1.5", -1.5), ("23.99999", 23.99999)],
)
def test_string_to_number(string, expected):
    assert string_to_number(string) == expected


@pytest.mark.parametrize("string", ["", "10000s", " 1.5", "-", ".5", "1e-4", "1.2.3"])
@pytest.mark.parametrize("required", [True, False])
def test_string_to_number_invalid(string, required):
    if required:
        with pytest.raises(ValueError):
            string_to_number(string)
    else:
        string_to_number(string, required=False) == string



from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools.command.sdist import sdist as _sdist
from distutils import log
import os
import sys
import pkg_resources
from io import open
import re


argv = sys.argv[1:]
needs_wheel = {"bdist_wheel"}.intersection(argv)
wheel = ["wheel"] if needs_wheel else []

# check if minimum required Cython is available
cython_version_re = re.compile(r'\s*"cython\s*>=\s*([0-9][0-9\w\.]*)\s*"')
with open("pyproject.toml", "r", encoding="utf-8") as fp:
    for line in fp:
        m = cython_version_re.match(line)
        if m:
            cython_min_version = m.group(1)
            break
    else:
        sys.exit("error: could not parse cython version from pyproject.toml")
try:
    required_cython = "cython >= %s" % cython_min_version
    pkg_resources.require(required_cython)
except pkg_resources.ResolutionError:
    with_cython = False
else:
    with_cython = True


class cython_build_ext(_build_ext):
    """Compile *.pyx source files to *.c using cythonize if Cython is
    installed, else use the pre-generated *.c sources.
    """

    def finalize_options(self):
        if with_cython:
            from Cython.Build import cythonize

            # optionally enable line tracing for test coverage support
            linetrace = os.environ.get("CYTHON_TRACE") == "1"

            self.distribution.ext_modules[:] = cythonize(
                self.distribution.ext_modules,
                force=linetrace or self.force,
                annotate=os.environ.get("CYTHON_ANNOTATE") == "1",
                quiet=not self.verbose,
                compiler_directives={
                    "linetrace": linetrace,
                    "language_level": 3,
                    "embedsignature": True,
                },
                include_path=["src"],
            )
        else:
            log.warn(
                "%s not installed; using pre-generated *.c sources" % required_cython
            )
            for ext in self.distribution.ext_modules:
                ext.sources = [re.sub(r"\.pyx$", ".c", n) for n in ext.sources]

        _build_ext.finalize_options(self)


class cython_sdist(_sdist):
    """Run 'cythonize' on *.pyx sources to ensure the *.c files included
    in the source distribution are up-to-date.
    """

    def run(self):
        if not with_cython:
            from distutils.errors import DistutilsSetupError

            raise DistutilsSetupError(
                "Cython >= %s is required to make sdist" % cython_min_version
            )

        from Cython.Build import cythonize

        cythonize(
            self.distribution.ext_modules,
            force=True,
            quiet=not self.verbose,
            compiler_directives={"language_level": 3, "embedsignature": True},
            include_path=["src"],
        )
        _sdist.run(self)


# need to include this for Visual Studio 2008 doesn't have stdint.h
include_dirs = (
    [os.path.join(os.path.dirname(__file__), "vendor", "msinttypes")]
    if os.name == "nt" and sys.version_info < (3,)
    else []
)

cython_modules = ["parser", "util", "writer", "_test"]
extensions = [
    Extension(
        "openstep_plist." + mod,
        sources=["src/openstep_plist/%s.pyx" % mod],
        include_dirs=include_dirs,
        language="c++",
        extra_compile_args=["-std=c++11"] if sys.platform != "win32" else [],
    )
    for mod in cython_modules
]

with open("README.md", "r") as f:
    long_description = f.read()

version_file = os.path.join("src", "openstep_plist", "_version.py")

setup_args = dict(
    name="openstep_plist",
    use_scm_version={"write_to": version_file},
    description="ASCII plist parser written in Cython",
    author="Cosimo Lupo",
    author_email="cosimo@anthrotype.com",
    url="https://github.com/fonttools/openstep-plist",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    ext_modules=extensions,
    setup_requires=["setuptools_scm"] + wheel,
    python_requires=">=3.8",
    cmdclass={"build_ext": cython_build_ext, "sdist": cython_sdist},
    zip_safe=False,
)


if __name__ == "__main__":
    setup(**setup_args)
